package cn.com.qmth.examcloud.core.oe.admin.api.provider;

import cn.com.qmth.examcloud.api.commons.enums.ExamType;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.core.oe.admin.api.SyncExamDataCloudService;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.*;
import cn.com.qmth.examcloud.core.oe.admin.api.request.SyncExamDataReq;
import cn.com.qmth.examcloud.core.oe.admin.api.response.SyncExamDataResp;
import cn.com.qmth.examcloud.core.oe.admin.dao.*;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.*;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.*;
import cn.com.qmth.examcloud.core.oe.admin.service.*;
import cn.com.qmth.examcloud.core.oe.admin.service.cache.ExamStudentCache;
import cn.com.qmth.examcloud.examwork.api.ExamSkipFaceCloudService;
import cn.com.qmth.examcloud.examwork.api.request.CheckExamSkipFaceReq;
import cn.com.qmth.examcloud.examwork.api.response.CheckExamSkipFaceResp;
import cn.com.qmth.examcloud.question.commons.core.question.AnswerType;
import cn.com.qmth.examcloud.support.CacheConstants;
import cn.com.qmth.examcloud.support.cache.CacheHelper;
import cn.com.qmth.examcloud.support.cache.bean.ExamPropertyCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExamSettingsCacheBean;
import cn.com.qmth.examcloud.support.enums.ExamProperties;
import cn.com.qmth.examcloud.support.helper.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import cn.com.qmth.examcloud.web.support.ControllerSupport;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Example;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

/**
 * @Description 同步考试相关数据接口
 * @Author lideyin
 * @Date 2019/12/20 10:32
 * @Version 1.0
 */
@Api(tags = "同步考试相关数据接口")
@RestController
@RequestMapping("${$rmp.cloud.oe}/exam/sync")
public class SyncExamDataCloudServiceProvider extends ControllerSupport implements SyncExamDataCloudService {

    private static final long serialVersionUID = -25466948667789119L;

    private static final Logger LOG = LoggerFactory.getLogger(SyncExamDataCloudServiceProvider.class);

    @Autowired
    private ExamRecordForMarkingService examRecordForMarkingService;

    @Autowired
    private ExamScorePushQueueService examScorePushQueueService;

    @Autowired
    private ExamScoreObtainQueueService examScoreObtainQueueService;

    @Autowired
    private ExamScoreNoticeQueueService examScoreNoticeQueueService;

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamRecordForMarkingRepo examRecordForMarkingRepo;

    @Autowired
    private ExamStudentRepo examStudentRepo;

    @Autowired
    private ExamSyncCaptureRepo examSyncCaptureRepo;

    @Autowired
    private ExamCaptureRepo examCaptureRepo;

    @Autowired
    private ExamFaceLivenessVerifyRepo examFaceLivenessVerifyRepo;

    @Autowired
    private ExamRecordPaperStructRepo examRecordPaperStructRepo;

    @Autowired
    private ExamRecordQuestionsRepo examRecordQuestionsRepo;

    @Autowired
    private FaceBiopsyRepo faceBiopsyRepo;

    @Autowired
    private FaceBiopsyItemRepo faceBiopsyItemRepo;

    @Autowired
    private FaceBiopsyItemStepRepo faceBiopsyItemStepRepo;

    @Autowired
    private ExamFaceLiveVerifyRepo examFaceLiveVerifyRepo;

    @Autowired
    private ExamAuditService examAuditService;

    @Autowired
    private ExamRecordDataSyncRepo examRecordDataSyncRepo;

    @Autowired
    private ExamScoreRepo examScoreRepo;

    @Autowired
    private ExamStudentCache examStudentCache;

    @Autowired
    private ExamStudentFinalScoreService examStudentFinalScoreService;

    @Autowired
    private ExamContinuedRecordRepo examContinuedRecordRepo;

    @Autowired
    private ExamProcessRecordRepo examProcessRecordRepo;

    @Autowired
    private ExamCaptureCameraInfoRepo examCaptureCameraInfoRepo;

    @Autowired
    private ExamWarnService examWarnService;

    @Autowired
    private ExamSkipFaceCloudService examSkipFaceCloudService;

    @Autowired
    private RedisClient redisClient;

    /**
     * 同步考试记录数据
     */
    @ApiOperation(value = "同步考试记录相关数据")
    @PostMapping("/syncExamData")
    @Transactional
    @Override
    public SyncExamDataResp syncExamData(@RequestBody SyncExamDataReq req) {
        long startTime = System.currentTimeMillis();

        //临时考试记录
        ExamRecordDataBean tempExamRecordData = req.getExamRecordData();

        //校验考试记录是否已同步，如果已同步，则直接返回
        if (hasSynced(tempExamRecordData.getId())) {
            return new SyncExamDataResp();
        }

        Long tempExamRecordDataId = tempExamRecordData.getId();
        Long examStudentId = tempExamRecordData.getExamStudentId();
        Long studentId = tempExamRecordData.getStudentId();
        Long examId = tempExamRecordData.getExamId();
        Long rootOrgId = tempExamRecordData.getRootOrgId();
        String examType = tempExamRecordData.getExamType();

        // 是否开启人脸验证
        boolean isFaceEnable = FaceBiopsyHelper.isFaceEnable(rootOrgId, examId, studentId);

        boolean skipFace = false, isFaceCheck = false, isFaceVerify = false, isFaceVerifyForceExit = false,
                isVirtualCameraAuditEnabled = false, isStrangerEnable = false;
        if (isFaceEnable) {
            // 检查当前考生是否跳过人脸识别
            skipFace = this.checkExamSkipFace(examId, examStudentId);

            // 是否开启考中人脸活体检测
            isFaceVerify = FaceBiopsyHelper.isFaceVerify(rootOrgId, examId, studentId);

            // 是否强制人脸验证
            isFaceCheck = FaceBiopsyHelper.isFaceCheck(examId, studentId);

            // 是否考中活体不通过强制退出
            ExamPropertyCacheBean faceVerifyForceExitProperty = CacheHelper.getExamProperty(examId, ExamProperties.FACE_VERIFY_FORCE_EXIT.name());
            isFaceVerifyForceExit = faceVerifyForceExitProperty != null && StringUtil.isTrue(faceVerifyForceExitProperty.getValue());

            // 是否开启虚拟摄像头进入待审核
            ExamPropertyCacheBean virtualCameraAuditEnabledProperty = CacheHelper.getExamProperty(examId, ExamProperties.VIRTUAL_CAMERA_AUDIT_ENABLED.name());
            isVirtualCameraAuditEnabled = virtualCameraAuditEnabledProperty != null && StringUtil.isTrue(virtualCameraAuditEnabledProperty.getValue());

            // 是否启用陌生人检测
            ExamPropertyCacheBean strangerProperty = CacheHelper.getExamProperty(examId, ExamProperties.IS_STRANGER_ENABLE.name());
            isStrangerEnable = strangerProperty != null && StringUtil.isTrue(strangerProperty.getValue());
        }

        // 是否开启审核全通过
        ExamPropertyCacheBean auditAllPassProperty = CacheHelper.getExamProperty(examId, ExamProperties.AUDIT_ALL_PASS.name());
        boolean isAuditAllPass = auditAllPassProperty != null && StringUtil.isTrue(auditAllPassProperty.getValue());

        // 是否推送分数
        ExamPropertyCacheBean pushScoreProperty = CacheHelper.getExamProperty(examId, ExamProperties.PUSH_SCORE.name());
        boolean isPushScore = pushScoreProperty != null && StringUtil.isTrue(pushScoreProperty.getValue());

        //先更新考生的考试次数
        ExamStudentEntity examStudent = updateExamStudent(examStudentId);

        //同步考试记录,并返回真实的考试记录id
        ExamRecordDataEntity realExamRecordData = syncExamRecordData(tempExamRecordData, examStudent);
        Long realExamRecordDataId = realExamRecordData.getId();

        //添加同步记录
        addExamRecordDataSync(tempExamRecordDataId, realExamRecordDataId);

        //同步考试记录对应的试卷结构
        syncExamRecordPaperStruct(req.getExamRecordPaperStruct(), realExamRecordData);
        //同步作答记录
        syncExamRecordQuestions(req.getExamRecordQuestions(), realExamRecordData);

        //同步考试过程记录表
        syncExamProcessRecords(req.getExamProcessRecords(), realExamRecordDataId);

        //同步断点续考记录表
        syncExamContinuedRecords(req.getExamContinuedRecords(), realExamRecordDataId);

        //同步考试分数表
        Long examScoreId = syncExamScore(tempExamRecordData, realExamRecordDataId);

        //同步“开考人脸识别验证”结果
        syncExamSyncCapture(req.getExamSyncCapture(), realExamRecordDataId);

        //同步“人脸抓拍比对”结果
        syncExamCapture(req.getExamCaptures(), realExamRecordDataId, isStrangerEnable);

        //同步活检结果：FACE_ID("S1", "FaceID活体检测方案")
        syncExamFaceLivenessVerify(req.getExamFaceLivenessVerifies(), realExamRecordDataId);

        //同步活检结果： FACE_MOTION("S2", "自研活体检测方案")【已废弃】
        syncFaceBiopsy(req.getFaceBiopsy(), realExamRecordDataId);

        //同步活检结果： FACE_CLIENT("S3", "C端活体检测方案")
        syncFaceLiveVerifyRecords(req.getFaceLiveVerifyRecords(), realExamRecordDataId);

        // ******************** 处理违纪、异常、审核等场景 start ********************

        //【违纪】非法考生端应用
        String illegalReason = redisClient.get(CacheConstants.CACHE_OE_DISCIPLINE_ILLEGAL_CLIENT + tempExamRecordDataId, String.class);
        if (StringUtils.isNotEmpty(illegalReason)) {
            examAuditService.saveExamAuditForIllegalClient(realExamRecordDataId, illegalReason);
            LOG.warn("{}_{}_{} ILLEGAL_CLIENT...", studentId, tempExamRecordDataId, realExamRecordDataId);
        }

        //【违纪】非法数据（已废弃）
        Boolean hasIllegal = redisClient.get(CacheConstants.CACHE_OE_DISCIPLINE_ILLEGAL_DATA + tempExamRecordDataId, Boolean.class);
        if (hasIllegal != null && hasIllegal) {
            examAuditService.saveHeaderWarnAudit(realExamRecordDataId);
            LOG.warn("{}_{}_{} ILLEGAL_DATA...", studentId, tempExamRecordDataId, realExamRecordDataId);
        }

        //【违纪】超过切屏次数限制
        if (tempExamRecordData.getExceedMaxSwitchScreenCount() != null && tempExamRecordData.getExceedMaxSwitchScreenCount()) {
            examAuditService.saveExceedMaxSwitchScreenCount(realExamRecordDataId);
            LOG.warn("{}_{}_{} EXCEED_MAX_SWITCH_SCREEN_COUNT...", studentId, tempExamRecordDataId, realExamRecordDataId);
        }

        //若开启人脸验证
        if (isFaceEnable) {

            //若强制人脸验证
            if (isFaceCheck && skipFace) {
                //【异常】人工跳过人脸验证
                examWarnService.saveExamWarn(realExamRecordDataId, WarnType.SKIP_FACE);
                //修改为异常状态
                examRecordDataRepo.updateExamRecordIsWarnById(realExamRecordDataId, true);
                LOG.warn("{}_{}_{} SKIP_FACE...", studentId, tempExamRecordDataId, realExamRecordDataId);
            }

            //无“开考人脸识别”结果
            if (req.getExamSyncCapture() == null) {
                //【异常】开考照片缺失
                examWarnService.saveExamWarn(realExamRecordDataId, WarnType.MISS_FACE);
                //修改为异常状态
                examRecordDataRepo.updateExamRecordIsWarnById(realExamRecordDataId, true);
                LOG.warn("{}_{}_{} MISS_FACE...", studentId, tempExamRecordDataId, realExamRecordDataId);
            }

            //无“人脸抓拍比对”结果
            if (CollectionUtils.isEmpty(req.getExamCaptures())) {
                //【违纪】抓拍照片数量为0
                examAuditService.saveExamAuditByNoPhoto(realExamRecordDataId);
                LOG.warn("{}_{}_{} OTHER_抓拍照片数量为0...", studentId, tempExamRecordDataId, realExamRecordDataId);
            } else {
                if (isVirtualCameraAuditEnabled) {
                    //存在虚拟摄像头，则进入待审
                    boolean hasVirtualCamera = req.getExamCaptures().stream().anyMatch(e -> e.getHasVirtualCamera() != null && e.getHasVirtualCamera());
                    if (hasVirtualCamera) {
                        //修改为异常状态
                        examRecordDataRepo.updateExamRecordIsWarnById(realExamRecordDataId, true);
                        LOG.warn("{}_{}_{} 存在虚拟摄像头...", studentId, tempExamRecordDataId, realExamRecordDataId);
                    }
                }
            }

            // 若开启考中人脸活体检测 且 活检失败
            if (isFaceVerify && IsSuccess.FAILED.name().equals(tempExamRecordData.getFaceVerifyResult())) {
                if (isFaceVerifyForceExit) {
                    //【违纪】指定动作失败
                    examAuditService.saveExamAuditByFaceVerifyFailed(realExamRecordDataId, rootOrgId);
                    LOG.warn("{}_{}_{} ACTION_FAILURE...", studentId, tempExamRecordDataId, realExamRecordDataId);
                } else {
                    //【异常】过程活体不通过
                    examWarnService.saveExamWarn(realExamRecordDataId, WarnType.FACE_LIVE_UNPASS);
                    //修改为异常状态
                    examRecordDataRepo.updateExamRecordIsWarnById(realExamRecordDataId, true);
                    LOG.warn("{}_{}_{} FACE_LIVE_UNPASS...", studentId, tempExamRecordDataId, realExamRecordDataId);
                }
            }
        }

        //若开启审核全通过
        if (isAuditAllPass) {
            auditPassExamRecordData(realExamRecordDataId);
            LOG.warn("{}_{}_{} AuditAllPass...", studentId, tempExamRecordDataId, realExamRecordDataId);
        }

        // ******************** 处理违纪、异常、审核等场景 end ********************

        //计算最终分数（这个步骤必须放在所有改变“违纪、警告、审核状态”的操作之后）
        if (ExamType.ONLINE.name().equals(examType) || ExamType.ONLINE_HOMEWORK.name().equals(examType)
                || ExamType.OFFLINE.name().equals(examType)) {
            examStudentFinalScoreService.calcAndSaveFinalScore(examStudentId);
        }

        // 保存阅卷相关数据
        examRecordForMarkingService.saveExamRecordForMarking(realExamRecordDataId, tempExamRecordData.getObjectiveScore());

        if (isPushScore) {
            // 保存考试分数数据到推分队列
            examScorePushQueueService.saveScoreDataInfoToQueue(realExamRecordDataId, examScoreId);

            // 保存考试分数数据到分数获取队列
            examScoreObtainQueueService.saveExamScoreObtainQueue(realExamRecordDataId);

            // 保存成绩通知队列
            examScoreNoticeQueueService.addExamScoreNoticeQueue(rootOrgId);
        }

        //刷新考生的缓存
        examStudentCache.refresh(examStudentId);

        LOG.warn("syncExamData finish, {}_{}_{} {}_{}_{}_{}_{}_{} cost:{}ms",
                studentId, tempExamRecordDataId, realExamRecordDataId, isFaceEnable, skipFace, isFaceVerify,
                isFaceCheck, isFaceVerifyForceExit, isAuditAllPass, System.currentTimeMillis() - startTime);

        return new SyncExamDataResp();
    }

    private void auditPassExamRecordData(Long realExamRecordDataId) {
        //开启审核全通过
        Optional<ExamRecordDataEntity> optional = examRecordDataRepo.findById(realExamRecordDataId);
        if (optional.isPresent()) {
            boolean isPass = examAuditService.updateExamAuditByAllPass(realExamRecordDataId);
            // 修改考试记录为已审
            ExamRecordDataEntity entity = optional.get();
            entity.setIsWarn(false);
            entity.setIsIllegality(false);
            if (isPass) {
                entity.setIsAudit(true);
            } else {
                entity.setIsAudit(false);
            }
            examRecordDataRepo.save(entity);
        }
    }

    private boolean checkExamSkipFace(Long examId, Long examStudentId) {
        // 检查当前考生是否跳过人脸识别
        CheckExamSkipFaceReq req = new CheckExamSkipFaceReq();
        req.setExamId(examId);
        req.setExamStudentId(examStudentId);
        CheckExamSkipFaceResp resp = examSkipFaceCloudService.checkExamSkipFace(req);
        return resp.isSkip();
    }

    private void syncExamContinuedRecords(List<ExamContinuedRecordBean> examContinuedRecords, Long realExamRecordDataId) {
        if (CollectionUtils.isEmpty(examContinuedRecords)) {
            return;
        }
        for (ExamContinuedRecordBean bean : examContinuedRecords) {
            ExamContinuedRecordEntity entity = copyExamContinuedRecordFrom(bean, realExamRecordDataId);
            examContinuedRecordRepo.saveAndFlush(entity);
        }
    }

    private void syncExamProcessRecords(List<ExamProcessRecordBean> examProcessRecords, Long realExamRecordDataId) {
        if (CollectionUtils.isEmpty(examProcessRecords)) {
            return;
        }
        for (ExamProcessRecordBean bean : examProcessRecords) {
            ExamProcessRecordEntity entity = copyExamProcessRecordFrom(bean, realExamRecordDataId);
            examProcessRecordRepo.saveAndFlush(entity);
        }
    }

    public Long syncExamScore(ExamRecordDataBean tempExamRecordData, Long realExamRecordDataId) {
        //如果考试成绩不存在，则新增，否则更新数据
        ExamScoreEntity examScoreEntity = examScoreRepo.findByExamRecordDataId(realExamRecordDataId);
        if (examScoreEntity == null) {
            examScoreEntity = new ExamScoreEntity();
            examScoreEntity.setExamRecordDataId(realExamRecordDataId);
        }
        examScoreEntity.setObjectiveScore(tempExamRecordData.getObjectiveScore());
        examScoreEntity.setObjectiveAccuracy(tempExamRecordData.getObjectiveAccuracy());
        examScoreEntity.setTotalScore(tempExamRecordData.getTotalScore());//交卷时,总分=客观分得分
        examScoreEntity.setSubjectiveScore(tempExamRecordData.getSubjectiveScore());
        examScoreEntity.setSuccPercent(tempExamRecordData.getSuccPercent());

        ExamScoreEntity result = examScoreRepo.saveAndFlush(examScoreEntity);
        return result.getId();
    }

    /**
     * 更新考生相关信息
     * 增加考试次数，修改考试状态为已完成
     */
    private ExamStudentEntity updateExamStudent(Long examStudentId) {
        //更新数据库中的已考次数
        ExamStudentEntity examStudentEntity = examStudentRepo.findByExamStudentId(examStudentId);
        if (null == examStudentEntity) {
            throw new StatusException("100101", "考生ID不正确！" + examStudentId);
        }

        int usedTimes = examStudentEntity.getUsedNum() == null ? 0 : examStudentEntity.getUsedNum();
        examStudentEntity.setUsedNum(usedTimes + 1);
        examStudentEntity.setFinished(true);

        examStudentRepo.saveAndFlush(examStudentEntity);
        return examStudentEntity;
    }

    /**
     * 添加考试记录的同步记录
     */
    private void addExamRecordDataSync(Long cacheId, Long dbId) {
        ExamRecordDataSyncEntity entity = new ExamRecordDataSyncEntity();
        entity.setCacheId(cacheId);
        entity.setDbId(dbId);

        examRecordDataSyncRepo.saveAndFlush(entity);
    }

    /**
     * 考试记录表是否已同步
     */
    private boolean hasSynced(Long cacheExamRecordDataId) {
        ExamRecordDataSyncEntity query = new ExamRecordDataSyncEntity();
        query.setCacheId(cacheExamRecordDataId);
        Example<ExamRecordDataSyncEntity> example = Example.of(query);
        return examRecordDataSyncRepo.exists(example);
    }

    private ExamRecordDataEntity syncExamRecordData(ExamRecordDataBean examRecordData, ExamStudentEntity examStudentEntity) {
        ExamRecordDataEntity result = new ExamRecordDataEntity();
        Long examId = examRecordData.getExamId();
        result.setExamId(examId);
        result.setExamType(ExamType.valueOf(examRecordData.getExamType()));
        result.setExamStudentId(examRecordData.getExamStudentId());
        Long studentId = examRecordData.getStudentId();
        result.setStudentId(studentId);

        //部分数据从考生表中获取
        result.setStudentCode(examStudentEntity.getStudentCode());
        result.setStudentName(examStudentEntity.getStudentName());
        result.setIdentityNumber(examStudentEntity.getIdentityNumber());
        result.setInfoCollector(examStudentEntity.getInfoCollector());

        // 已考次数
        Integer usedNum = examStudentEntity.getUsedNum();
        // 允许考试次数
        ExamSettingsCacheBean examCache = ExamCacheTransferHelper.getDefaultCachedExam(examId);
        int examTimes = examCache.getExamTimes() == null ? 0 : examCache.getExamTimes().intValue();
        // 第几次考试
        result.setExamOrder(getExamOrder(examTimes, usedNum));
        // 是否重考
        boolean isReexamine = isReexamine(examTimes, usedNum);
        result.setIsReexamine(isReexamine);
        //如果有重考，则需要保存重考相关数据
        if (isReexamine) {
            result.setReexamineType(examStudentEntity.getReexamineType());
            result.setReexamineDetail(examStudentEntity.getReexamineDetail());
            this.handleReexamine(examRecordData.getExamStudentId(), usedNum, examTimes, examStudentEntity.getExtraNum());
        }

        result.setCourseId(examRecordData.getCourseId());
        result.setCourseLevel(examStudentEntity.getCourseLevel());

        result.setOrgId(examRecordData.getOrgId());
        result.setRootOrgId(examRecordData.getRootOrgId());
        result.setBasePaperId(examRecordData.getBasePaperId());

        result.setRandomPaper(examRecordData.getRandomPaper());
        result.setPaperScore(examRecordData.getPaperScore());

        // result.setPaperStructId(examRecordData.getPaperStructId());
        result.setPaperType(examRecordData.getPaperType());

        result.setStartTime(examRecordData.getStartTime() != null ? examRecordData.getStartTime() : examRecordData.getEnterExamTime());
        result.setEndTime(examRecordData.getEndTime());
        result.setCleanTime(examRecordData.getCleanTime());
        result.setIsWarn(examRecordData.getWarn());
        result.setIsAudit(examRecordData.getAudit());
        result.setIsIllegality(examRecordData.getIllegality());
        result.setExamRecordStatus(ExamRecordStatus.getByName(examRecordData.getExamRecordStatus()));
        result.setUsedExamTime(examRecordData.getUsedExamTime());

        result.setIsContinued(examRecordData.getContinued());
        result.setContinuedCount(examRecordData.getContinuedCount());
        result.setFaceSuccessCount(examRecordData.getFaceSuccessCount());
        result.setFaceFailedCount(examRecordData.getFaceFailedCount());
        result.setFaceStrangerCount(examRecordData.getFaceStrangerCount());
        result.setFaceTotalCount(examRecordData.getFaceTotalCount());
        result.setFaceSuccessPercent(examRecordData.getFaceSuccessPercent());

        if (StringUtils.isNotBlank(examRecordData.getFaceVerifyResult())) {
            result.setFaceVerifyResult(IsSuccess.valueOf(examRecordData.getFaceVerifyResult()));
        }

        result.setFaceLandmarkVal(examRecordData.getFaceLandmarkVal());
        result.setIsAllObjectivePaper(examRecordData.getAllObjectivePaper());
        result.setBaiduFaceLivenessSuccessPercent(examRecordData.getBaiduFaceLivenessSuccessPercent());
        result.setIsExceed(examRecordData.getExceed());
        result.setExamStageId(examRecordData.getExamStageId());
        result.setExamStageOrder(examRecordData.getExamStageOrder());
        result.setLastActiveTime(examRecordData.getLastActiveTime());
        result.setEnterExamTime(examRecordData.getEnterExamTime());
        result.setContinuedTime(examRecordData.getContinuedTime());
        result.setSwitchScreenCount(examRecordData.getSwitchScreenCount());
        if (result.getSwitchScreenCount() == null) {
            result.setSwitchScreenCount(0);
        }

        return examRecordDataRepo.saveAndFlush(result);
    }

    /**
     * 处理重考
     */
    private void handleReexamine(Long examStudentId, Integer usedNum, Integer examTimes, Integer extraNum) {
        List<ExamRecordDataEntity> examRecords = examRecordDataRepo.findByExamStudentId(examStudentId);
        Optional<ExamRecordDataEntity> examRecordOptional = examRecords.stream()
                .filter(examRecordData -> examRecordData.getExamRecordStatus() != ExamRecordStatus.EXAM_INVALID
                        && examRecordData.getIsReexamine() != null && examRecordData.getIsReexamine())
                .findFirst();

        if (examRecordOptional.isPresent()) {
            ExamRecordDataEntity examRecordData = examRecordOptional.get();

            // 将上一次重考的记录设置成无效
            examRecordData.setExamRecordStatus(ExamRecordStatus.EXAM_INVALID);
            examRecordDataRepo.save(examRecordData);

            // 删除生成的阅卷数据，避免传到阅卷
            ExamRecordForMarkingEntity examRecordForMarking = examRecordForMarkingRepo.findByExamRecordDataId(examRecordData.getId());
            if (examRecordForMarking != null) {
                examRecordForMarkingRepo.delete(examRecordForMarking);
            }

            LOG.warn("处理重考，将上一次重考记录作废！examId:{} studentId:{} examStudentId:{} examRecordDataId:{} usedNum:{} examTimes:{} extraNum:{}",
                    examRecordData.getExamId(), examRecordData.getStudentId(), examStudentId, examRecordData.getId(),
                    usedNum, examTimes, extraNum);
        }
    }

    /**
     * 计算考试次数
     */
    private Integer getExamOrder(Integer examTimes, Integer usedNum) {
        // 超过可考次数,始终为可考次数+1
        if (usedNum > examTimes) {
            return examTimes + 1;
        }
        return usedNum;
    }

    /**
     * 计算是否为重考
     */
    private boolean isReexamine(Integer examTimes, Integer usedNum) {
        // 超过可考次数,则认为有重考
        if (usedNum > examTimes) {
            return true;
        }
        return false;
    }

    private void syncExamRecordQuestions(ExamRecordQuestionsBean examRecordQuestions, ExamRecordDataEntity examRecordData) {
        ExamRecordQuestionsEntity entity = copyExamRecordQuestionsFrom(examRecordQuestions, examRecordData.getId());
        examRecordQuestionsRepo.save(entity);

        //更新考试记录表中的作答记录id
        examRecordData.setExamRecordQuestionsId(entity.getId());
        examRecordDataRepo.save(examRecordData);
    }

    private ExamRecordQuestionsEntity copyExamRecordQuestionsFrom(ExamRecordQuestionsBean bean, Long realExamRecordDataId) {
        ExamRecordQuestionsEntity entity = new ExamRecordQuestionsEntity();
        entity.setExamRecordDataId(realExamRecordDataId);

        List<ExamQuestionEntity> examQuestionEntityList = new ArrayList<>();

        for (ExamQuestionBean questionBean : bean.getExamQuestionBeans()) {
            ExamQuestionEntity questionEntity = new ExamQuestionEntity();

            questionEntity.setExamRecordDataId(realExamRecordDataId);
            questionEntity.setMainNumber(questionBean.getMainNumber());
            questionEntity.setQuestionId(questionBean.getQuestionId());
            questionEntity.setOrder(questionBean.getOrder());
            questionEntity.setStudentAnswer(questionBean.getStudentAnswer());
            questionEntity.setStudentScore(questionBean.getStudentScore());
            questionEntity.setQuestionScore(questionBean.getQuestionScore());
            questionEntity.setQuestionType(questionBean.getQuestionType());
            questionEntity.setIsAnswer(questionBean.getAnswer());
            questionEntity.setIsSign(questionBean.getSign());
            questionEntity.setCorrectAnswer(questionBean.getCorrectAnswer());
            questionEntity.setOptionPermutation(questionBean.getOptionPermutation());
            questionEntity.setAudioPlayTimes(questionBean.getAudioPlayTimes());
            if (null != questionBean.getAnswerType()) {
                questionEntity.setAnswerType(AnswerType.valueOf(questionBean.getAnswerType()));
            }

            examQuestionEntityList.add(questionEntity);
        }

        entity.setExamQuestionEntities(examQuestionEntityList);
        return entity;
    }

    private void syncExamRecordPaperStruct(ExamRecordPaperStructBean examRecordPaperStruct, ExamRecordDataEntity examRecordData) {
        ExamRecordPaperStructEntity entity = copyExamRecordPaperStructFrom(examRecordPaperStruct);
        examRecordPaperStructRepo.save(entity);

        //更新考试记录表中的试卷结构id
        examRecordData.setPaperStructId(entity.getId());
        examRecordDataRepo.save(examRecordData);
    }

    private ExamRecordPaperStructEntity copyExamRecordPaperStructFrom(ExamRecordPaperStructBean examRecordPaperStruct) {
        ExamRecordPaperStructEntity entity = new ExamRecordPaperStructEntity();
        entity.setId(examRecordPaperStruct.getId());
        entity.setDefaultPaper(examRecordPaperStruct.getDefaultPaper());
        return entity;
    }

    @Transactional
    public void syncFaceBiopsy(FaceBiopsyBean faceBiopsy, Long examRecordDataId) {
        if (faceBiopsy == null) {
            return;
        }
        FaceBiopsyEntity faceBiopsyEntity = copyFaceBiopsyFrom(faceBiopsy, examRecordDataId);
        faceBiopsyRepo.saveAndFlush(faceBiopsyEntity);

        for (FaceBiopsyItemBean itemBean : faceBiopsy.getFaceBiopsyItems()) {
            FaceBiopsyItemEntity faceBiopsyItemEntity = copyFaceBiopsyItemFrom(
                    itemBean, faceBiopsyEntity.getId(), examRecordDataId);
            faceBiopsyItemRepo.saveAndFlush(faceBiopsyItemEntity);

            List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList = copyFaceBiopsyItemStepFrom(
                    itemBean.getFaceBiopsyItemSteps(), faceBiopsyItemEntity.getId(), examRecordDataId);
            for (FaceBiopsyItemStepEntity fbis : faceBiopsyItemStepEntityList) {
                faceBiopsyItemStepRepo.saveAndFlush(fbis);
            }
        }
    }

    private void syncFaceLiveVerifyRecords(List<FaceLiveVerifyBean> faceLiveVerifyRecords, Long examRecordDataId) {
        if (CollectionUtils.isEmpty(faceLiveVerifyRecords)) {
            return;
        }

        for (FaceLiveVerifyBean bean : faceLiveVerifyRecords) {
            ExamFaceLiveVerifyEntity entity = new ExamFaceLiveVerifyEntity();

            entity.setExamRecordDataId(examRecordDataId);
            entity.setStatus(FaceLiveVerifyStatus.valueOf(bean.getStatus()));
            entity.setFaceCount(bean.getFaceCount());
            entity.setSimilarity(bean.getSimilarity());
            entity.setRealness(bean.getRealness());
            entity.setErrorMsg(bean.getErrorMsg());
            entity.setProcessTime(bean.getProcessTime());
            entity.setActions(bean.getActions());
            entity.setCreationTime(bean.getCreationTime());
            entity.setUpdateTime(bean.getUpdateTime());

            examFaceLiveVerifyRepo.save(entity);
        }
    }

    private FaceBiopsyEntity copyFaceBiopsyFrom(FaceBiopsyBean bean, Long examRecordDataId) {
        FaceBiopsyEntity entity = new FaceBiopsyEntity();
        entity.setRootOrgId(bean.getRootOrgId());
        entity.setExamRecordDataId(examRecordDataId);
        entity.setResult(bean.getResult());
        entity.setErrorMsg(bean.getErrorMsg());
        entity.setVerifiedTimes(bean.getVerifiedTimes());

        return entity;
    }

    private FaceBiopsyItemEntity copyFaceBiopsyItemFrom(FaceBiopsyItemBean bean, Long faceByiopsyId, Long examRecordDataId) {
        FaceBiopsyItemEntity entity = new FaceBiopsyItemEntity();
        entity.setExamRecordDataId(examRecordDataId);
        entity.setFaceBiopsyId(faceByiopsyId);
        entity.setFaceBiopsyType(FaceBiopsyType.valueOf(bean.getFaceBiopsyType()));
        entity.setCompleted(bean.getCompleted());
        entity.setResult(bean.getResult());
        entity.setErrorMsg(bean.getErrorMsg());
        entity.setInFreezeTime(bean.getInFreezeTime());

        return entity;
    }

    private List<FaceBiopsyItemStepEntity> copyFaceBiopsyItemStepFrom(
            List<FaceBiopsyItemStepBean> faceBiopsyItemSteps,
            Long faceBiopsyItemId, Long examRecordDataId) {
        List<FaceBiopsyItemStepEntity> faceBiopsyItemStepEntityList = new ArrayList<>();
        for (FaceBiopsyItemStepBean bean : faceBiopsyItemSteps) {
            FaceBiopsyItemStepEntity entity = new FaceBiopsyItemStepEntity();
            entity.setExamRecordDataId(examRecordDataId);
            entity.setFaceBiopsyItemId(faceBiopsyItemId);
            entity.setAction(FaceBiopsyAction.valueOf(bean.getAction()));
            entity.setResourceRelativePath(bean.getResourceRelativePath());
            entity.setResourceType(ResourceType.valueOf(bean.getResourceType()));
            entity.setActionStay(bean.getActionStay());
            entity.setResult(bean.getResult());
            entity.setErrorMsg(bean.getErrorMsg());
            entity.setExt1(bean.getExt1());
            entity.setExt2(bean.getExt2());
            entity.setExt3(bean.getExt3());
            entity.setExt4(bean.getExt4());
            entity.setExt5(bean.getExt5());

            faceBiopsyItemStepEntityList.add(entity);
        }

        return faceBiopsyItemStepEntityList;
    }

    private void syncExamFaceLivenessVerify(List<ExamFaceLivenessVerifyBean> verifyBeans, Long examRecordDataId) {
        if (CollectionUtils.isEmpty(verifyBeans)) {
            return;
        }
        for (ExamFaceLivenessVerifyBean bean : verifyBeans) {
            ExamFaceLivenessVerifyEntity entity = copyExamFaceLivenessVerifyFrom(bean, examRecordDataId);
            examFaceLivenessVerifyRepo.saveAndFlush(entity);
        }
    }

    private ExamFaceLivenessVerifyEntity copyExamFaceLivenessVerifyFrom(ExamFaceLivenessVerifyBean bean, Long examRecordDataId) {
        ExamFaceLivenessVerifyEntity entity = new ExamFaceLivenessVerifyEntity();
        entity.setExamRecordDataId(examRecordDataId);
        entity.setStartTime(bean.getStartTime());
        entity.setUsedTime(bean.getUsedTime());
        entity.setResultJson(bean.getResultJson());

        if (null != bean.getVerifyResult()) {
            entity.setVerifyResult(FaceVerifyResult.valueOf(bean.getVerifyResult()));
        }

        entity.setBizId(bean.getBizId());
        entity.setIsError(bean.getIsError());
        entity.setErrorMsg(bean.getErrorMsg());
        entity.setOperateNum(bean.getOperateNum());

        return entity;
    }

    private void syncExamSyncCapture(ExamSyncCaptureBean examSyncCapture, Long examRecordDataId) {
        if (examSyncCapture == null) {
            return;
        }
        ExamSyncCaptureEntity examSyncCaptureEntity = copyExamSyncCaptureFrom(examSyncCapture, examRecordDataId);
        examSyncCaptureRepo.saveAndFlush(examSyncCaptureEntity);
    }

    private ExamSyncCaptureEntity copyExamSyncCaptureFrom(ExamSyncCaptureBean examSyncCapture, Long examRecordDataId) {
        ExamSyncCaptureEntity entity = new ExamSyncCaptureEntity();
        entity.setExamRecordDataId(examRecordDataId);
        entity.setFileUrl(examSyncCapture.getFileUrl());
        entity.setFileName(examSyncCapture.getFileName());
        entity.setFaceCompareResult(examSyncCapture.getFaceCompareResult());
        entity.setIsPass(examSyncCapture.getPass());
        entity.setIsStranger(examSyncCapture.getStranger());
        entity.setLandmark(examSyncCapture.getLandmark());
        entity.setUsedTime(examSyncCapture.getUsedTime());
        entity.setFacelivenessResult(examSyncCapture.getFacelivenessResult());
        entity.setProcessTime(examSyncCapture.getProcessTime());
        entity.setHasVirtualCamera(examSyncCapture.getHasVirtualCamera());
        entity.setCameraInfos(examSyncCapture.getCameraInfos());
        entity.setExtMsg(examSyncCapture.getExtMsg());

        return entity;
    }

    private void syncExamCapture(List<ExamCaptureBean> examCaptures, Long examRecordDataId, boolean isStrangerEnable) {
        if (CollectionUtils.isEmpty(examCaptures)) {
            return;
        }

        for (ExamCaptureBean bean : examCaptures) {
            ExamCaptureEntity entity = new ExamCaptureEntity();
            entity.setExamRecordDataId(examRecordDataId);
            entity.setFileUrl(bean.getFileUrl());
            entity.setFileName(bean.getFileName());
            entity.setFaceCompareResult(bean.getFaceCompareResult());
            entity.setIsPass(bean.getPass());
            entity.setLandmark(bean.getLandmark());
            entity.setUsedTime(bean.getUsedTime());
            entity.setFacelivenessResult(bean.getFacelivenessResult());
            entity.setProcessTime(bean.getProcessTime());
            entity.setHasVirtualCamera(bean.getHasVirtualCamera());
            entity.setExtMsg(bean.getExtMsg());
            entity.setCameraInfos(null);
            if (isStrangerEnable) {
                entity.setIsStranger(bean.getStranger());
            } else {
                entity.setIsStranger(false);
            }

            saveExamCaptureCameraInfo(examRecordDataId, bean.getCameraInfos());

            examCaptureRepo.saveAndFlush(entity);
        }
    }

    private void saveExamCaptureCameraInfo(Long examRecordDataId, String cameraInfos) {
        if (StringUtils.isEmpty(cameraInfos)) {
            return;
        }
        try {
            JSONArray jsonArray = new JSONArray(cameraInfos);
            for (int i = 0; i < jsonArray.length(); i++) {
                JSONObject jsonObject = (JSONObject) jsonArray.get(i);
                ExamCaptureCameraInfoEntity ci = new ExamCaptureCameraInfoEntity();
                ci.setExamRecordDataId(examRecordDataId);
                ci.setDetail(getVal(jsonObject, "detail", 500));
                ci.setName(getVal(jsonObject, "name", 100));
                ci.setPid(getVal(jsonObject, "pid", 100));
                ci.setVid(getVal(jsonObject, "vid", 100));
                if (jsonObject.isNull("pid") || StringUtils.isBlank(jsonObject.getString("pid"))
                        || jsonObject.isNull("vid") || StringUtils.isBlank(jsonObject.getString("vid"))) {
                    ci.setVirtualCamera(true);
                } else {
                    ci.setVirtualCamera(false);
                }

                List<ExamCaptureCameraInfoEntity> olds = examCaptureCameraInfoRepo.findByExamRecordDataIdAndPidAndVidAndName(
                        examRecordDataId, ci.getPid(), ci.getVid(), ci.getName());
                if (CollectionUtils.isEmpty(olds)) {
                    examCaptureCameraInfoRepo.saveAndFlush(ci);
                } else {
                    ExamCaptureCameraInfoEntity old = olds.get(0);
                    if (!StringUtils.equals(old.getDetail(), ci.getDetail())) {
                        old.setDetail(ci.getDetail());
                        examCaptureCameraInfoRepo.saveAndFlush(old);
                    }
                }
            }
        } catch (Exception e) {
            LOG.error("保存摄像信息失败:" + examRecordDataId + " " + cameraInfos, e);
        }
    }

    private String getVal(JSONObject jsonObject, String name, int length) throws JSONException {
        if (jsonObject.isNull(name) || StringUtils.isBlank(jsonObject.getString(name))) {
            return "";
        } else {
            String content = jsonObject.getString(name);
            if (content.length() > length) {
                return content.substring(0, length);
            } else {
                return content;
            }
        }
    }

    private ExamContinuedRecordEntity copyExamContinuedRecordFrom(ExamContinuedRecordBean bean, Long examRecordDataId) {
        ExamContinuedRecordEntity entity = new ExamContinuedRecordEntity();
        entity.setExamRecordDataId(examRecordDataId);
        entity.setContinuedTime(bean.getContinuedTime());
        entity.setStartTime(bean.getStartTime());
        entity.setUsedExamTime(bean.getUsedExamTime());
        return entity;
    }

    private ExamProcessRecordEntity copyExamProcessRecordFrom(ExamProcessRecordBean bean, Long examRecordDataId) {
        ExamProcessRecordEntity entity = new ExamProcessRecordEntity();
        entity.setExamRecordDataId(examRecordDataId);
        entity.setProcessName(bean.getProcessName());
        entity.setRecordTime(bean.getRecordTime());
        entity.setSourceIp(bean.getSourceIp());
        return entity;
    }

}
